import { template, getNextElement, spread, effect, runHydrationEvents } from 'solid-js/web';
import { createSignal, createMemo, createEffect, onCleanup, on, splitProps, mergeProps } from 'solid-js';
import { EditorState, StateEffect } from '@codemirror/state';
import { basicSetup } from '@codemirror/basic-setup';
import { EditorView, keymap, placeholder } from '@codemirror/view';
import { indentWithTab } from '@codemirror/commands';
import { createStore } from 'solid-js/store';

const updateListenerExtension = (onChange) => {
  return EditorView.updateListener.of((vu) => {
    if (vu.docChanged && typeof onChange === "function") {
      const document = vu.state.doc;
      const value = document.toString();
      onChange(value, vu);
    }
  });
};

function createCodeMirror(initialOptions) {
  const [options, setOptions] = createStore(initialOptions);
  const [container, setContainer] = createSignal();
  const [view, setView] = createSignal();
  const [state, setState] = createSignal();
  const defaultThemeOption = EditorView.theme({
    "&": {
      height: initialOptions.height,
      minHeight: initialOptions.minHeight,
      maxHeight: initialOptions.maxHeight,
      width: initialOptions.width,
      minWidth: initialOptions.minWidth,
      maxWidth: initialOptions.maxWidth
    }
  });
  const computedExtensions = createMemo(() => {
    const {
      indentWithTab: indentWithTab$1,
      basicSetup: basicSetup$1,
      placeholder: placeholder$1,
      editable,
      onUpdate,
      extensions
    } = options;
    const updateListener = updateListenerExtension(options.onChange);
    return [
      indentWithTab$1 ? keymap.of([indentWithTab]) : null,
      basicSetup$1 ? basicSetup : null,
      placeholder$1 && placeholder ? placeholder(placeholder$1) : null,
      !editable ? EditorView.editable.of(false) : null,
      !!onUpdate && typeof onUpdate === "function" ? EditorView.updateListener.of(onUpdate) : null,
      updateListener,
      defaultThemeOption,
      ...extensions || null
    ].filter(Boolean);
  }, options);
  createEffect(() => {
    const { value, selection, root } = options;
    if (container() && !state()) {
      const stateCurrent = EditorState.create({
        doc: value,
        selection,
        extensions: computedExtensions()
      });
      setState(stateCurrent);
      if (!view()) {
        const viewCurrent = new EditorView({
          state: stateCurrent,
          parent: container(),
          root
        });
        setView(viewCurrent);
      }
    }
  });
  onCleanup(() => view()?.destroy());
  createEffect(() => {
    if (view()) {
      const currentValue = view()?.state.doc.toString();
      if (options.value !== currentValue) {
        view()?.dispatch({
          changes: {
            from: 0,
            to: currentValue?.length,
            insert: options.value || ""
          }
        });
      }
    }
  });
  createEffect(on(computedExtensions, () => view()?.dispatch({
    effects: StateEffect.reconfigure.of(computedExtensions())
  })));
  createEffect(() => {
    if (options.autoFocus && view()) {
      view()?.focus();
    }
  });
  return {
    options,
    setOptions,
    state,
    setState,
    view,
    setView,
    container,
    setContainer
  };
}

const extractCodeMirrorProps = (props) => {
  const [
    options,
    htmlAttrs
  ] = splitProps(props, [
    "className",
    "value",
    "selection",
    "onChange",
    "onUpdate",
    "autoFocus",
    "height",
    "minHeight",
    "maxHeight",
    "width",
    "minWidth",
    "maxWidth",
    "basicSetup",
    "placeholder",
    "editable",
    "root",
    "extensions",
    "indentWithTab"
  ]);
  const computedOptions = mergeProps({
    className: "",
    extensions: [],
    height: "auto",
    minHeight: "none",
    maxHeight: "none",
    placeHolder: "",
    width: "auto",
    minWidth: "none",
    maxWidth: "none",
    editable: true,
    indentWithTab: true,
    basicSetup: true
  }, options);
  return { options: computedOptions, htmlAttrs };
};

const _tmpl$ = template(`<div></div>`, 2);
const CodeMirror = props => {
  let editor;
  const {
    options,
    htmlAttrs
  } = extractCodeMirrorProps(props);
  const {
    view,
    setOptions,
    setContainer
  } = createCodeMirror({
    container: editor,
    ...options
  });
  createEffect(() => setContainer(editor));
  createEffect(() => setOptions(options));
  onCleanup(() => view()?.destroy());
  return (() => {
    const _el$ = getNextElement(_tmpl$);

    (ref => editor = ref)(_el$);

    spread(_el$, htmlAttrs, false, false);

    effect(() => _el$.className = `solid-cm ${options?.className ? ` ${options.className}` : ''}`);

    runHydrationEvents();

    return _el$;
  })();
};

export { CodeMirror, createCodeMirror };
